% Riemannian gradient algorithm for BTD with and without variable
% projection.
% This script builds an exact BTD and then tries to recover it using the
% Riemannian gradient algorithm with (btd_grad) and without (btd_grad_full)
% variable projection starting from the same starting iterates.
itermax_grad = 1e4;
itermax_grad_full = 2e4;
tol = 5e-14;
tol_obj = 1e-25;
%% Building of the BTD
I = [5 5 5];
R = 2;
Rank = [2 2 2];
A = zeros(I);
S = randn([Rank R]);
U = zeros(I(1),Rank(1),R);
V = zeros(I(2),Rank(2),R);
W = zeros(I(3),Rank(3),R);
for r = 1:R
    U(:,:,r) = qf(randn(I(1),Rank(1)));
    V(:,:,r) = qf(randn(I(2),Rank(2)));
    W(:,:,r) = qf(randn(I(3),Rank(3)));
    A = A + tmprod(S(:,:,:,r),{U(:,:,r),V(:,:,r),W(:,:,r)},[1 2 3]);
end
%% Recovering the BTD
sample = 100; % number of starting iterates
% Column 1 = grad, Column 2 = grad_full
success = zeros(2,1); % success means that both normgrad < tol and objective < tol_obj
normgrad = zeros(sample,2); % norm of the gradient at the end of the process
objective = zeros(sample,2); % objective at the end of the process
iter = zeros(sample,2); % number of iterations needed to converge
inner_iter = zeros(sample,2); % mean number of iterations in the backtracking loop
time = zeros(sample,3); % Column 3 = time spent in variable projection
for s = 1:sample
    % Initial iterate
    S0 = randn([Rank R]);
    U0 = zeros(I(1),Rank(1),R);
    V0 = zeros(I(2),Rank(2),R);
    W0 = zeros(I(3),Rank(3),R);
    for r = 1:R
        U0(:,:,r) = qf(randn(I(1),Rank(1)));
        V0(:,:,r) = qf(randn(I(2),Rank(2)));
        W0(:,:,r) = qf(randn(I(3),Rank(3)));
    end
    % grad
    start = tic;
    [~,~,~,~,info_grad] = btd_grad(A,U0,V0,W0,tol,itermax_grad);
    time(s,1) = toc(start);
    normgrad_grad = info_grad.normgrad;
    obj_grad = info_grad.objective;
    % grad_full
    start = tic;
    [~,~,~,~,info_grad_full] = btd_grad_full(A,S0,U0,V0,W0,tol,itermax_grad_full);
    time(s,2) = toc(start);
    normgrad_grad_full = info_grad_full.normgrad;
    obj_grad_full = info_grad_full.objective;
    % Infos storage
    normgrad(s,1) = normgrad_grad(end);
    normgrad(s,2) = normgrad_grad_full(end);
    objective(s,1) = obj_grad(end);
    objective(s,2) = obj_grad_full(end);
    iter(s,1) = length(normgrad_grad);
    iter(s,2) = length(normgrad_grad_full);
    inner_iter(s,1) = mean(info_grad.iterations);
    inner_iter(s,2) = mean(info_grad_full.iterations);
    time(s,3) = info_grad.VarProjTime;
    % Success counter
    if normgrad(s,1) <= tol && objective(s,1) <= tol_obj
        success(1,1) = success(1,1)+1;
        U0_grad = U0;
        V0_grad = V0;
        W0_grad = W0;
    end
    if normgrad(s,2) <= tol && objective(s,2) <= tol_obj
        success(2,1) = success(2,1)+1;
        U0_grad_full = U0;
        V0_grad_full = V0;
        W0_grad_full = W0;
    end
end